import argparse
import os
import sys
from getpass import getpass
from pathlib import Path

from steampunk_scanner import api
from steampunk_scanner.helpers import AnsibleEntity, parse_ansible_entities


def add_parser(subparsers):
    parser = subparsers.add_parser(
        "scan", help="Initiate Ansible scan", description="Initiate Ansible scan"
    )
    parser.add_argument(
        "--tasks", "-t", type=lambda p: Path(p).absolute(), nargs='+', help="Paths to file with Ansible tasks"
    )
    parser.add_argument(
        "--playbooks", "-p", type=lambda p: Path(p).absolute(), nargs='+', help="Paths to Ansible playbook file"
    )
    parser.add_argument(
        "--roles", "-r", type=lambda p: Path(p).absolute(), nargs='+', help="Paths to Ansible role directory"
    )
    parser.add_argument(
        "--collections", "-c", type=lambda p: Path(p).absolute(), nargs='+', help="Paths to Ansible collection"
    )
    parser.add_argument(
        "--output", "-o", type=argparse.FileType("w"), help="Output file location", default="-"
    )
    parser.set_defaults(func=_parser_callback)


def _parser_callback(args: argparse.Namespace):
    """
    Invoke the Ansible scanner and print/save the scan result
    :param args: Argparse arguments
    """
    username = os.environ.get("SCANNER_USERNAME") or input("Username: ")
    password = os.environ.get("SCANNER_PASSWORD") or getpass()
    client = api.Client(api.ENDPOINT, username, password)

    input_tasks = []
    if args.tasks:
        input_tasks += parse_ansible_entities(args.tasks, AnsibleEntity.TASK)
    if args.playbooks:
        input_tasks += parse_ansible_entities(args.playbooks, AnsibleEntity.PLAYBOOK)
    if args.roles:
        input_tasks += parse_ansible_entities(args.roles, AnsibleEntity.ROLE)
    if args.collections:
        input_tasks += parse_ansible_entities(args.collections, AnsibleEntity.COLLECTION)

    response = client.post("/scan-tasks", input_tasks)
    if response.ok:
        _print_scan_output(args.output, input_tasks, response.json())
    else:
        print(f"API error: {response.status_code} - {response.json()['msg']}")
        sys.exit(1)


def _print_scan_output(out_fh, input_tasks, output_tasks):
    """
    Prints scan output
    :param out_fh: File handle to print result to
    :param input_tasks: Input Ansible task list
    :param output_tasks: Output Ansible task list
    """
    for input_task, output_task in zip(input_tasks, output_tasks):
        file_name = input_task.get("__file__", None)
        task_line = input_task.get("__line__", None)
        certified = output_task.get("certified", False)
        errors = output_task.get("errors", [])
        fqcn = output_task.get("fqcn", None)
        hints = output_task.get("hints", [])

        if fqcn and not certified:
            print(
                f"{file_name}:{task_line}: WARNING: The {fqcn} module is not certified.",
                file=out_fh
            )

        for error in errors:
            print(f"{file_name}:{task_line}: ERROR: {error}", file=out_fh)

        for hint in hints:
            print(f"{file_name}:{task_line}: HINT: {hint}", file=out_fh)

    if errors:
        sys.exit(1)



import requests
import os

ENDPOINT = os.environ.get("SCANNER_ENDPOINT", "https://scanner.steampunk.si/api")


class Client:
    def __init__(self, url, username=None, password=None):
        self.url = url.rstrip("/")
        self.username = username
        self.password = password

    def post(self, path, payload):
        args = dict(json=payload)
        if self.username:
            args["auth"] = (self.username, self.password)
        return requests.post(f"{self.url}{path}", timeout=10, **args)



import argparse
from getpass import getpass

from steampunk_scanner import api


def add_parser(subparsers):
    parser = subparsers.add_parser(
        "account", help="Manage user account", description="Manage user account"
    )

    subsubparsers = parser.add_subparsers()
    parser_new = subsubparsers.add_parser(
        "register", help="Register a new account", description="Register a new account"
    )
    parser_new.add_argument(
        "username", type=str, help="User email"
    )
    parser_new.set_defaults(func=_parser_callback_new)

    parser_activate = subsubparsers.add_parser(
        "activate", help="Activate pending account", description="Activate pending account"
    )
    parser_activate.add_argument(
        "username", type=str, help="User email"
    )
    parser_activate.add_argument(
        "verification_code", type=str, help="User account verification code from email"
    )
    parser_activate.set_defaults(func=_parser_callback_activate)


def _parser_callback_new(args: argparse.Namespace):
    """
    Registers a new account
    :param args: Argparse arguments
    """
    password = getpass()
    client = api.Client(api.ENDPOINT)
    response = client.post(
        "/accounts/register", dict(username=args.username, password=password)
    )
    if response.ok:
        print(response.json()["msg"])
    else:
        print(f"API error: {response.status_code} - {response.json()['msg']}")
        exit(1)


def _parser_callback_activate(args: argparse.Namespace):
    """
    Activates pending account
    :param args: Argparse arguments
    """
    client = api.Client(api.ENDPOINT)
    response = client.post(
        "/accounts/activate",
        dict(username=args.username, verification_code=args.verification_code)
    )
    if response.ok:
        print(response.json()["msg"])
    else:
        print(f"API error: {response.status_code} - {response.json()['msg']}")
        exit(1)



import argparse
import inspect
import sys

from steampunk_scanner import commands


class ArgParser(argparse.ArgumentParser):
    """An argument parser that displays help on error"""

    def error(self, message):
        """
        Overridden the original error method
        :param message: Error message
        """
        sys.stderr.write("error: {}\n".format(message))
        self.print_help()
        sys.exit(2)

    def add_subparsers(self, **kwargs):
        """
        Overridden the original add_subparsers method (workaround for http://bugs.python.org/issue9253)
        """
        subparsers = super(ArgParser, self).add_subparsers()
        subparsers.required = True
        subparsers.dest = "command"
        return subparsers


def create_parser():
    """
    Create argument parser for CLI
    :return: Parser as argparse.ArgumentParser object
    """
    parser = ArgParser(description="Steampunk Scanner - a quality scanner for Ansible Playbooks")

    subparsers = parser.add_subparsers()
    cmds = inspect.getmembers(commands, inspect.ismodule)
    for _, module in sorted(cmds, key=lambda x: x[0]):
        module.add_parser(subparsers)
    return parser


def main():
    """
    Main CLI method to be called
    """
    parser = create_parser()
    args = parser.parse_args()
    return args.func(args)



import sys
from enum import Enum
from pathlib import Path
from typing import List, Optional, Union

import yaml


class SafeLineLoader(yaml.loader.SafeLoader):
    """YAML loader that adds line numbers"""

    def construct_mapping(self, node, deep=False) -> dict:
        """
        Overridden the original construct_mapping method
        :param node: YAML node object
        :param deep: Build objects recursively
        :return: A dict with loaded YAML
        """
        mapping = super(SafeLineLoader, self).construct_mapping(node, deep=deep)
        mapping['__line__'] = node.start_mark.line + 1
        return mapping


class AnsibleEntity(Enum):
    """Enum that can distinct between different Ansible entities"""
    TASK = 1
    PLAYBOOK = 2
    ROLE = 3
    COLLECTION = 4


def _parse_tasks(tasks: List, file_name: str, collections: Optional[List] = None) -> List:
    """
    Parse Ansible tasks and prepare them for scanning
    :param tasks: List of Ansible task dicts
    :param file_name: Name of the original file with tasks
    :param collections: List of Ansible collections for tasks
    :return: List of parsed Ansible tasks that are prepared for scanning
    """
    for task in tasks:
        task.pop("name", None)
        task_line = task.pop("__line__", None)

        # TODO: Remove this spaghetti when API will be able to parse action plugins
        if "action" in task:
            dict_with_module = next((d for d in list(task.values()) if type(d) is dict and "module" in d), None)
            if dict_with_module is not None:
                module_name = dict_with_module.pop("module", None)
                action = task.pop("action", None)
                task[module_name] = action

        for task_key in task:
            if type(task[task_key]) is dict:
                task[task_key].pop("__line__", None)
                task[task_key] = list(task[task_key].keys())
            else:
                task[task_key] = None

        if collections is None:
            collections = []

        task["collections"] = collections
        task["__file__"] = file_name
        task["__line__"] = task_line

    return tasks


def _parse_playbook(playbook: dict, file_name: str) -> List:
    """
    Parse Ansible playbook and prepare it for scanning
    :param playbook: Ansible playbook as dict
    :param file_name: Name of the original file with playbook
    :return: List of parsed Ansible tasks that are prepared for scanning
    """
    playbook.pop("name", None)
    playbook.pop("hosts", None)
    playbook.pop("__line__", None)
    collections = playbook.pop("collections", None)

    return _parse_tasks(playbook.get("tasks", []), file_name, collections)


def _parse_task_file(file: Path) -> List:
    """
    Parse Ansible file
    :param file: Ansible task file
    :return: List of parsed Ansible tasks that are prepared for scanning
    """
    with file.open() as stream:
        return _parse_tasks(yaml.load(stream, Loader=SafeLineLoader), str(file))


def _parse_playbook_file(file: Path) -> List:
    """
    Parse Ansible playbook file
    :param file: Ansible playbook file
    :return: List of parsed Ansible tasks that are prepared for scanning
    """
    with file.open() as stream:
        return _parse_playbook(yaml.load(stream, Loader=SafeLineLoader)[0], str(file.relative_to(Path.cwd())))


def _parse_role_dir(directory: Path) -> List:
    """
    Parse Ansible role
    :param directory: Role directory
    :return: List of parsed Ansible tasks that are prepared for scanning
    """
    parsed_role = []
    for file in (list((directory / "tasks").rglob("*")) + list((directory / "handlers").rglob("*"))):
        if file.is_file():
            with file.open() as stream:
                yaml_dict = yaml.load(stream, Loader=SafeLineLoader)
                parsed_tasks = _parse_tasks(yaml_dict, str(file))
                parsed_role += parsed_tasks
    return parsed_role


def parse_role_dirs(dirs: List[Path]) -> List:
    """
    Parse multiple Ansible roles
    :param dirs: List of roles as directories
    :return: List of parsed Ansible tasks that are prepared for scanning
    """
    parsed_files = []
    for directory in dirs:
        if not directory.is_dir():
            print(f"File {directory.name} is not a valid directory")
            sys.exit(1)
        parsed_files += _parse_role_dir(directory)

    return parsed_files


def _parse_collection_dir(directory: Path) -> List:
    """
    Parse Ansible collection
    :param directory: Collection directory
    :return: List of parsed Ansible tasks that are prepared for scanning
    """
    parsed_collection = []
    for role in (list((directory / "roles").rglob("*"))):
        if role.is_dir():
            parsed_collection += _parse_role_dir(role)
    for playbook in (list((directory / "playbooks").rglob("*"))):
        if playbook.is_file():
            parsed_collection += _parse_playbook_file(playbook)
    return parsed_collection


def parse_ansible_entities(paths: Union[Path, List[Path]], ansible_entity_type: AnsibleEntity) -> List:
    """
    Parse multiple Ansible entities
    :param paths: Path or list of paths to Ansible entities
    :param ansible_entity_type: Type of Ansible files (task files, playbooks, roles or collections)
    :return: List of parsed Ansible tasks that are prepared for scanning
    """
    parsed_ansible_entities = []
    if isinstance(paths, list):
        for path in paths:
            if ansible_entity_type == AnsibleEntity.TASK:
                if not path.is_file():
                    print(f"Task file {path.name} is not a valid file")
                    sys.exit(1)
                parsed_ansible_entities += _parse_task_file(path)
            if ansible_entity_type == AnsibleEntity.PLAYBOOK:
                if not path.is_file():
                    print(f"Playbook {path.name} is not a valid file")
                    sys.exit(1)
                parsed_ansible_entities += _parse_playbook_file(path)
            if ansible_entity_type == AnsibleEntity.ROLE:
                if not path.is_dir():
                    print(f"Role {path.name} is not a valid directory")
                    sys.exit(1)
                parsed_ansible_entities += _parse_role_dir(path)
            if ansible_entity_type == AnsibleEntity.COLLECTION:
                if not path.is_dir():
                    print(f"Collection {path.name} is not a valid directory")
                    sys.exit(1)
                parsed_ansible_entities += _parse_collection_dir(path)

    return parsed_ansible_entities



#!/usr/bin/env python

import setuptools


# constructs the local component of the version without the commit hash
def local_scheme(version):
    return ""


setuptools.setup(use_scm_version={"local_scheme": local_scheme})
